#!/usr/bin/env python

"""NOT a good example of twisted-x11 code, but a shiny demo and stresstest.

This started as a pretty direct port of the default xcompmgr mode.
Any bugs added are my own.

There are various problems with this code:

- The xcompmgr code (obviously) assumes functions run sequentially.
  That does not happen here: functions can interleave at yield calls.
  I have made pretty much no checks whether that makes sense, other
  than fixing up the obvious failures I actually hit.

  (an example: Window.init contained some yields before w.attr gets set.
  If we get a CreateNotifyEvent followed by a MapNotifyEvent we ended up
  setting w.attr.map_state from the second event handler before the first
  one got around to initializing w.attr, which doesn't work very well.)

- Error handling is sloppy in various ways:

  - It relies heavily on t.i.defer.setDebugging. An underdocumented
    trick is that you actually get *better* debugging output with that
    on and deferred errors unhandled than with an explicit
    addErrback(log.err).

  - Many things we do are inherently racy. We (obviously) operate on
    windows we do not own, and do not always cope if they go away at
    the wrong time. We also do not necessarily cope with some events
    or redraws happening in parallel.
"""

from __future__ import absolute_import, division

from twisted.python import log
from twisted.internet import defer

from tx11 import core, event, renderutil
from tx11.atoms import XA_ATOM, XA_CARDINAL, XA_PIXMAP
from tx11.proto import xproto, render, damage, composite, xfixes


# XXX this is missing from xfixes.xml in xcb-proto for some reason...
# Snarfed it from xfixeswire.h for now.

from tx11.base import Enum

class WindowRegion(Enum):
    Bounding = 0
    Clip = 1


BACKGROUND_PROPS = ['_XROOTPMAP_ID', '_XSETROOT_ID']

ATOM_NAMES = ['_NET_WM_CM_S0',
              '_NET_WM_WINDOW_OPACITY',
              '_NET_WM_WINDOW_TYPE',
              '_NET_WM_WINDOW_TYPE_DESKTOP',
              '_NET_WM_WINDOW_TYPE_DOCK',
              '_NET_WM_WINDOW_TYPE_TOOLBAR',
              '_NET_WM_WINDOW_TYPE_MENU',
              '_NET_WM_WINDOW_TYPE_UTILITY',
              '_NET_WM_WINDOW_TYPE_SPLASH',
              '_NET_WM_WINDOW_TYPE_DIALOG',
              '_NET_WM_WINDOW_TYPE_NORMAL',
              ] + BACKGROUND_PROPS

OPAQUE = 0xffffffff
DEFAULT_OPACITY = OPAQUE # or 0xcccccccc while you experiment with this

# XXX there is currently some race if I set this lower. Needs to die though.
PAINT_DELAY = 0.01

WINDOW_SOLID, WINDOW_TRANS, WINDOW_ARGB = xrange(3)


def startup(reactor):
    d = core.connect(reactor,
                     extensions=[render.RenderExtension,
                                 damage.DamageExtension,
                                 composite.CompositeExtension,
                                 xfixes.XFixesExtension,
                                 ])
    compositor = Compositor()
    d.addCallback(compositor.run)
    return d


@defer.inlineCallbacks
def _get_wintype_atom(conn, wid):
    prop = yield conn.core.getProperty(False, wid,
                                       conn.atoms['_NET_WM_WINDOW_TYPE'],
                                       XA_ATOM, 0, 1)
    normal = conn.atoms['_NET_WM_WINDOW_TYPE_NORMAL']
    if not prop.value:
        wtype = normal
    elif len(prop.value) > 1:
        raise NotImplementedError(repr(prop.value))
    else:
        wtype = prop.value[0]

    if wtype != normal:
        defer.returnValue(wtype)

    query = yield conn.core.queryTree(wid)
    childtypes = [_get_wintype_atom(conn, child) for child in query.children]
    for childtype in (yield defer.gatherResults(childtypes)):
        if childtype != normal:
            defer.returnValue(wtype)

    defer.returnValue(wtype)


class Window(object):

    def __init__(self, compositor, window, x, y, width, height, border_width):
        self.compositor = compositor
        self.conn = compositor.conn
        self.id = window
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.border_width = border_width

    @defer.inlineCallbacks
    def init(self):
        self.damaged = False
        self.damage = None
        self.pixmap = None
        self.picture = self.alphaPict = None
        self.borderSize = None
        self.extents = None
        self.opacity = DEFAULT_OPACITY

        self.borderClip = None

        attrs = self.conn.core.getWindowAttributes(self.id)
        wtype = _get_wintype_atom(self.conn, self.id)

        self.attrs = yield attrs
        # Why do we do this instead of just using .attrs directly?
        # Because this function races with further events.
        # We may get a call to map() before the yield above returns,
        # which wants to update the map_state.
        self.map_state = self.attrs.map_state
        self.override_redirect = self.attrs.override_redirect
        self.windowtype = yield wtype

        if self.attrs.class_ != xproto.WindowClass.InputOnly:
            self.damage = yield self.conn.generateId()
            self.conn.damage.create(self.damage, self.id,
                                    damage.ReportLevel.NonEmpty)

        if self.map_state == xproto.MapState.Viewable:
            yield self.map()

    def destroy(self):
        if self.picture is not None:
            self.conn.render.freePicture(self.picture)
            self.picture = None
        if self.alphaPict is not None:
            self.conn.render.freePicture(self.alphaPict)
            self.alphaPict = None
        if self.damage is not None:
            d = self.conn.damage.destroy(self.damage)
            # XXX I have no idea if this is sane...
            d.addErrback(lambda f: f.trap(damage.BadDamage))
            self.damage = None

    @defer.inlineCallbacks
    def map(self):
        self.map_state = xproto.MapState.Viewable
        # "This needs to be here or else we lose transparency messages"
        d = self.conn.core.changeWindowAttributes(
            self.id, eventMask=xproto.EventMask.PropertyChange)
        # Apparently we do hit this for windows that are already gone...
        # XXX not 100% sure this is sane
        d.addErrback(lambda f: f.trap(xproto.BadWindow))
        # "This needs to be here since we don't get PropertyNotify
        # when unmapped"
        self.opacity = yield self._get_opacity_prop(DEFAULT_OPACITY)
        yield self._determine_mode()

        self.damaged = False

    def unmap(self):
        self.damaged = False
        if self.extents is not None:
            # "destroys region"
            self.compositor.add_damage(self.extents)
            self.extents = None
        if self.pixmap is not None:
            self.conn.core.freePixmap(self.pixmap)
            self.pixmap = None

        if self.picture is not None:
            self.conn.render.freePicture(self.picture)
            self.picture = None

        # "Don't care about properties anymore"
        d = self.conn.core.changeWindowAttributes(self.id, eventMask=0)
        # We do this for windows that have been destroyed, so
        # BadWindow is a nonissue.
        d.addErrback(lambda f: f.trap(xproto.BadWindow))

        if self.borderSize is not None:
            self.conn.xfixes.destroyRegion(self.borderSize)
            self.borderSize = None

        if self.borderClip is not None:
            self.conn.xfixes.destroyRegion(self.borderClip)
            self.borderClip = None

        self.compositor.clipChanged = True

    @defer.inlineCallbacks
    def _get_opacity_prop(self, default):
        prop = yield self.conn.core.getProperty(
            False, self.id, self.conn.atoms['_NET_WM_WINDOW_OPACITY'],
            XA_CARDINAL, 0, 1)
        if not prop.value:
            defer.returnValue(default)
        if len(prop.value) > 1:
            raise NotImplementedError(repr(prop.value))
        defer.returnValue(prop.value[0])

    @defer.inlineCallbacks
    def _determine_mode(self):
        if self.alphaPict is not None:
            self.conn.render.freePicture(self.alphaPict)
            self.alphaPict = None
        if self.attrs.class_ == xproto.WindowClass.InputOnly:
            format = None
        else:
            formid = renderutil.find_visual_format(
                self.conn.pictformats.screens[self.conn.preferredScreen],
                None, self.attrs.visual)
            for format in self.conn.pictformats.formats:
                if format.id == formid:
                    break
            else:
                raise LookupError()

        if format is not None \
                and format.type == render.PictType.Direct \
                and format.direct.alpha_mask:
            self.mode = WINDOW_ARGB
        elif self.opacity != OPAQUE:
            self.mode = WINDOW_TRANS
        else:
            self.mode = WINDOW_SOLID

        if self.extents is not None:
            damage = yield self.conn.generateId()
            # XXX iiuc this is daft: we should have an add_damage that
            # does not destroy the region passed in so we can pass
            # extents directly.
            self.conn.xfixes.createRegion(damage, [])
            self.conn.xfixes.copyRegion(self.extents, damage)
            self.compositor.add_damage(damage)

    @defer.inlineCallbacks
    def getExtents(self):
        x = self.x
        y = self.y
        width = self.width + 2 * self.border_width
        height = self.height + 2 * self.border_width
        # TODO: If we were doing shadows we would include their area here.
        region = yield self.conn.generateId()
        self.conn.xfixes.createRegion(region, [[x, y, width, height]])
        defer.returnValue(region)


class Compositor(object):

    @defer.inlineCallbacks
    def run(self, conn):
        self.root = root = conn.setup.roots[conn.preferredScreen]
        self.conn = conn
        self.windows = {}
        self.windowstack = []
        self.rootBuffer = None
        self.rootTile = None
        self.expose_rects = []
        self.paintTimer = None
        self.allDamage = None
        self.clipChanged = True

        # a bit roundtrip-happy, improve this later?

        conn.eventHandler = handler = event.SynchronousHandler()
        handler.addEventHandler(self.handleCreate, xproto.CreateNotifyEvent)
        handler.addEventHandler(self.handleConfigure,
                                xproto.ConfigureNotifyEvent)
        handler.addEventHandler(self.handleDestroy, xproto.DestroyNotifyEvent)
        handler.addEventHandler(self.handleMap, xproto.MapNotifyEvent)
        handler.addEventHandler(self.handleUnmap, xproto.UnmapNotifyEvent)
        handler.addEventHandler(self.handleReparent,
                                xproto.ReparentNotifyEvent)
        handler.addEventHandler(self.handleCirculate,
                                xproto.CirculateNotifyEvent)
        handler.addEventHandler(self.handleExpose, xproto.ExposeEvent)
        handler.addEventHandler(self.handleProp, xproto.PropertyNotifyEvent)
        handler.addEventHandler(self.handleDamage, damage.NotifyEvent)

        # Note these are actually required: the Damage extension at least
        # does not work until you've told it your client-side version.
        ver_comp, ver_damage, ver_fixes = yield defer.gatherResults([
                conn.composite.queryVersion(composite.MAJOR_VERSION,
                                            composite.MINOR_VERSION),
                conn.damage.queryVersion(damage.MAJOR_VERSION,
                                         damage.MINOR_VERSION),
                conn.xfixes.queryVersion(xfixes.MAJOR_VERSION,
                                         xfixes.MINOR_VERSION),
                ])

        if ver_comp.major_version != 0 or ver_comp.minor_version < 3:
            print >> sys.stderr, 'unsupported composite version %s.%s' % (
                ver_comp.major_version, ver_comp.minor_version)
            defer.returnValue(None)

        if ver_damage.major_version != damage.MAJOR_VERSION or \
                ver_damage.minor_version < 1:
            print >> sys.stderr, 'unsupported damage version %s.%s' % (
                ver_damage.major_version, ver_damage.minor_version)
            defer.returnValue(None)

        if ver_fixes.major_version != xfixes.MAJOR_VERSION or \
                ver_fixes.minor_version < xfixes.MINOR_VERSION:
            print >> sys.stderr, 'unsupported xfixes version %s.%s' % (
                ver_fixes.major_version, ver_fixes.minor_version)
            defer.returnValue(None)

        window = yield conn.generateId()
        conn.core.createWindow(
            root.root_depth, window, root.root, 0, 0, 1, 1, 0,
            xproto.WindowClass.InputOutput, root.root_visual)

        # TODO: wm properties

        atoms = yield defer.gatherResults([conn.core.internAtom(False, s)
                                           for s in ATOM_NAMES])
        conn.atoms = dict((name, atom.atom)
                          for name, atom in zip(ATOM_NAMES, atoms))

        # TODO: don't hardcode screen number (0?)
        conn.core.setSelectionOwner(window, conn.atoms['_NET_WM_CM_S0'], 0)

        reply = self.conn.pictformats = yield conn.render.queryPictFormats()
        self._formatARGB32 = renderutil.find_format(
            reply, renderutil.STANDARD_ARGB_32)
        self._formatRGB24 = renderutil.find_format(
            reply, renderutil.STANDARD_RGB_24)
        self._formatA8 = renderutil.find_format(
            reply, renderutil.STANDARD_A_8)
        self._formatVisual = renderutil.find_visual_format(
            reply.screens[conn.preferredScreen],
            root.root_depth, root.root_visual)

        self.rootPicture = yield conn.generateId()
        conn.render.createPicture(
            self.rootPicture, root.root, self._formatVisual,
            subwindowMode=xproto.SubwindowMode.IncludeInferiors)

        blackPicture = yield self.solid_picture(root, True, 1, 0, 0, 0)

        log.msg('acquiring initial grab...')
        conn.core.grabServer()
        try:
            query = yield conn.core.queryTree(root.root)
            geoms = yield defer.gatherResults([conn.core.getGeometry(window)
                                               for window in query.children])
            for child, geom in reversed(zip(query.children, geoms)):
                self.windows[child] = window = Window(
                    self, child, geom.x, geom.y, geom.width, geom.height,
                    geom.border_width)
                self.windowstack.append(window)
            yield defer.gatherResults([win.init()
                                       for win in self.windowstack])
            conn.composite.redirectSubwindows(root.root,
                                              composite.Redirect.Manual)
            conn.core.changeWindowAttributes(root.root, eventMask=(
                    xproto.EventMask.SubstructureNotify |
                    xproto.EventMask.Exposure |
                    xproto.EventMask.StructureNotify |
                    xproto.EventMask.PropertyChange))
        finally:
            log.msg('releasing server grab...')
            conn.core.ungrabServer()

        region = yield conn.generateId()
        conn.xfixes.createRegion(region, [[0, 0,
                                           self.root.width_in_pixels,
                                           self.root.height_in_pixels]])
        self.paint_all(region)

    @defer.inlineCallbacks
    def solid_picture(self, root, argb, a, r, g, b):
        pixmap = yield self.conn.generateId()
        depth = 32 if argb else 8
        self.conn.core.createPixmap(depth, pixmap, root.root, 1, 1)
        pid = yield self.conn.generateId()

        format = self._formatARGB32 if argb else self._formatA8
        # XXX xcompmgr does ca.repeat = True...
        self.conn.render.createPicture(
            pid, pixmap, format.id, repeat=render.Repeat.Normal)

        color = [int(x) for x in [r * 65535, g * 65535, b * 65535, a * 65535]]
        self.conn.render.fillRectangles(render.PictOp.Src, pid, color,
                                        [[0, 0, 1, 1]])
        self.conn.core.freePixmap(pixmap)
        defer.returnValue(pid)

    @defer.inlineCallbacks
    def paint_all(self, region):
        conn = self.conn
        root = self.root
        if self.rootBuffer is None:
            rootPixmap = yield conn.generateId()
            conn.core.createPixmap(root.root_depth, rootPixmap, root.root,
                                   root.width_in_pixels, root.height_in_pixels
                                   )
            self.rootBuffer = yield conn.generateId()
            conn.render.createPicture(self.rootBuffer, rootPixmap,
                                      self._formatVisual)
            conn.core.freePixmap(rootPixmap)

        conn.xfixes.setPictureClipRegion(self.rootPicture, region, 0, 0)
        # We iterate over this twice, once forwards, once backwards.
        windows = []
        for window in self.windowstack:
            if not window.damaged:
                # "never painted, ignore it"
                continue
            # "if invisible, ignore it"
            if window.x + window.width < 1 \
                    or window.y + window.height < 1 \
                    or window.x >= self.root.width_in_pixels \
                    or window.y >= self.root.height_in_pixels:
                continue
            if window.picture is None:
                if window.pixmap is None:
                    window.pixmap = yield conn.generateId()
                    conn.composite.nameWindowPixmap(window.id, window.pixmap)
                format = renderutil.find_visual_format(
                    conn.pictformats.screens[conn.preferredScreen], None,
                    window.attrs.visual)
                window.picture = yield conn.generateId()
                conn.render.createPicture(
                    window.picture, window.pixmap, format,
                    subwindowMode=xproto.SubwindowMode.IncludeInferiors)
            if self.clipChanged:
                if window.borderSize is not None:
                    conn.xfixes.destroyRegion(window.borderSize)
                    window.borderSize = None
                if window.extents is not None:
                    conn.xfixes.destroyRegion(window.extents)
                    window.extents = None
                if window.borderClip is not None:
                    conn.xfixes.destroyRegion(window.borderClip)
                    window.borderClip = None
            if window.borderSize is None:
                # "if window doesn't exist anymore, this will generate
                # an error as well as not generate a region. Perhaps a
                # better XFixes architecture would be to have a request
                # that copies instead of creates, that way you'd just
                # end up with an empty region instead of an invalid XID."
                window.borderSize = yield conn.generateId()
                conn.xfixes.createRegionFromWindow(window.borderSize,
                                                   window.id,
                                                   WindowRegion.Bounding)
                conn.xfixes.translateRegion(
                    window.borderSize,
                    window.x + window.border_width,
                    window.y + window.border_width)
            if window.extents is None:
                window.extents = yield window.getExtents()
            if window.mode == WINDOW_SOLID:
                x = window.x
                y = window.y
                width = window.width + 2 * window.border_width
                height = window.height + 2 * window.border_width
                conn.xfixes.setPictureClipRegion(self.rootBuffer, region,
                                                 0, 0)
                # XXX I hope I got the order right here
                conn.xfixes.subtractRegion(region, region, window.borderSize)
                conn.render.composite(render.PictOp.Src, window.picture,
                                      0, self.rootBuffer, 0, 0, 0, 0,
                                      x, y, width, height)
            if window.borderClip is None:
                window.borderClip = yield conn.generateId()
                conn.xfixes.createRegion(window.borderClip, [])
                conn.xfixes.copyRegion(region, window.borderClip)
            windows.append(window)

        conn.xfixes.setPictureClipRegion(self.rootBuffer, region, 0, 0)

        if self.rootTile is None:
            pixmap = None
            # XXX this is broken currently
            for propname in []: #BACKGROUND_PROPS:
                prop = yield conn.core.getProperty(
                    False, self.root.root, conn.atoms[propname],
                    xproto.GetPropertyType.Any, 0, 4)
                if prop.type == XA_PIXMAP and prop.format == 32:
                    if len(prop.value) != 1:
                        raise NotImplementedError(repr(prop.value))
                    pixmap = prop.value[0]
                    fill = True
                    break
            if pixmap is None:
                pixmap = yield conn.generateId()
                conn.core.createPixmap(root.root_depth, pixmap,
                                       root.root, 1, 1)
                fill = True
            self.rootTile = yield conn.generateId()
            # XXX xcompmgr does ca.repeat = True...
            conn.render.createPicture(self.rootTile, pixmap,
                                      self._formatVisual,
                                      repeat=render.Repeat.Normal)
            if fill:
                conn.render.fillRectangles(
                    render.PictOp.Src, self.rootTile,
                    [0x8080, 0x8080, 0x8080, 0xffff],
                    [[0, 0, 1, 1]])

        conn.render.composite(render.PictOp.Src,
                              self.rootTile, 0, self.rootBuffer,
                              0, 0, 0, 0, 0, 0,
                              self.root.width_in_pixels,
                              self.root.height_in_pixels)

        for window in reversed(windows):
            conn.xfixes.setPictureClipRegion(self.rootBuffer,
                                             window.borderClip, 0, 0)
            # TODO: If we were drawing shadows we would do so here
            if window.opacity != OPAQUE and window.alphaPict is None:
                window.alphaPict = yield self.solid_picture(
                    self.root, False, window.opacity / OPAQUE, 0, 0, 0)
            # XXX how does this work in xcompmgr?
            # If I read it right w->alphaPict is only set above
            # (if opacity != OPAQUE), but it is passed to XRenderComposite
            # for WINDOW_ARGB. Am I missing a spot where opacity != OPAQUE
            # for argb windows, or is that error just ignored?
            # Actually it looks like None is a valid value...
            if window.mode in (WINDOW_TRANS, WINDOW_ARGB):
                x = window.x
                y = window.y
                width = window.width + window.border_width * 2
                height = (window.height +
                          window.border_width * 2)
                mask = window.alphaPict
                if mask is None:
                    mask = 0
                conn.render.composite(
                    render.PictOp.Over,
                    window.picture, mask, self.rootBuffer,
                    0, 0, 0, 0, x, y, width, height)
            conn.xfixes.destroyRegion(window.borderClip)
            window.borderClip = None
        conn.xfixes.destroyRegion(region)
        conn.xfixes.setPictureClipRegion(self.rootBuffer, 0, 0, 0)
        conn.render.composite(render.PictOp.Src,
                              self.rootBuffer, 0, self.rootPicture,
                              0, 0, 0, 0, 0, 0,
                              self.root.width_in_pixels,
                              self.root.height_in_pixels)

    def handleCreate(self, conn, event):
        self.windows[event.window] = window = Window(
            self, event.window,
            event.x, event.y, event.width, event.height, event.border_width)
        self.windowstack.append(window)
        window.init()

    @defer.inlineCallbacks
    def handleConfigure(self, conn, event):
        try:
            window = self.windows[event.window]
        except KeyError:
            if event.window == self.root.root:
                if self.rootBuffer is not None:
                    conn.render.freePicture(self.rootBuffer)
                    rootBuffer = None
                # XXX HACK
                self.root.width_in_pixels = event.width
                self.root.height_in_pixels = event.heigh
            # XXX do we hit this?
            log.msg('configure event for unknown window?')
            defer.returnValue(None)

        damage = yield conn.generateId()
        conn.xfixes.createRegion(damage, [])
        if window.extents is not None:
            conn.xfixes.copyRegion(window.extents, damage)

        window.x = event.x
        window.y = event.y

        if window.width != event.width \
                or window.height != event.height:
            if window.pixmap is not None:
                conn.core.freePixmap(window.pixmap)
                window.pixmap = None
                if window.picture is not None:
                    conn.render.freePicture(window.picture)
                    window.picture = None
            window.width = event.width
            window.height = event.height

        window.border_width = event.border_width
        window.override_redirect = event.override_redirect
        if event.above_sibling:
            try:
                above = self.windows[event.above_sibling]
            except KeyError:
                log.msg('unhandled above?')
        else:
            above = None
        self.restack(window, above)

        extents = yield window.getExtents()
        conn.xfixes.unionRegion(damage, extents, damage)
        conn.xfixes.destroyRegion(extents)

        self.add_damage(damage)
        self.clipChanged = True

    def handleDestroy(self, conn, event):
        try:
            window = self.windows[event.window]
        except KeyError:
            log.msg('destroy window unhandled?')
        else:
            window.unmap()
            window.destroy()
            del self.windows[event.window]
            self.windowstack.remove(window)

    def handleMap(self, conn, event):
        try:
            window = self.windows[event.window]
        except KeyError:
            log.msg('map window unhandled?')
        else:
            window.map()

    def handleUnmap(self, conn, event):
        try:
            window = self.windows[event.window]
        except KeyError:
            log.msg('unmap window unhandled?')
        else:
            window.unmap()

    def handleReparent(self, conn, event):
        try:
            window = self.windows[event.window]
        except KeyError:
            log.msg('reparent window unhandled?')
        else:
            if event.parent == self.root.root:
                self.handleCreate(conn, event)
            else:
                window.destroy()
                del self.windows[event.window]
                self.windowstack.remove(window)

    def handleCirculate(self, conn, event):
        try:
            window = self.windows[event.window]
        except KeyError:
            log.msg('circulate unhandled?')
        else:
            if event.place == xproto.Place.OnTop:
                new_above = self.topWindow
            else:
                new_above = None
            self.restack(window, new_above)

    @defer.inlineCallbacks
    def handleExpose(self, conn, event):
        if event.window == self.root.root:
            self.expose_rects.append([event.x, event.y,
                                      event.width, event.height])
            if not event.count:
                region = yield conn.generateId()
                conn.xfixes.createRegion(region, self.expose_rects)
                self.add_damage(region)
                self.expose_rects = []

    @defer.inlineCallbacks
    def handleProp(self, conn, event):
        for propname in BACKGROUND_PROPS:
            if event.atom == conn.atoms[propname]:
                if self.rootTile is not None:
                    conn.core.clearArea(True, self.root.root, 0, 0, 0, 0)
                    conn.render.freePicture(self.rootTile)
                    self.rootTile = None
                    break
        if event.atom == conn.atoms['_NET_WM_WINDOW_OPACITY']:
            window = self.windows[event.window]
            window.opacity = yield window._get_opacity_prop(DEFAULT_OPACITY)
            window._determine_mode()

    @defer.inlineCallbacks
    def handleDamage(self, conn, event):
        try:
            window = self.windows[event.drawable]
        except KeyError:
            log.msg('unhandled damage?')
            return

        if not window.damaged:
            parts = yield window.getExtents()
            conn.damage.subtract(window.damage, 0, 0)
        else:
            parts = yield conn.generateId()
            conn.xfixes.createRegion(parts, [])
            conn.damage.subtract(window.damage, 0, parts)
            conn.xfixes.translateRegion(
                parts,
                window.x + window.border_width,
                window.y + window.border_width)
        self.add_damage(parts)
        window.damaged = True

    def add_damage(self, damage):
        if self.allDamage is not None:
            self.conn.xfixes.unionRegion(self.allDamage, damage,
                                         self.allDamage)
            self.conn.xfixes.destroyRegion(damage)
        else:
            self.allDamage = damage

        if self.paintTimer is None:
            from twisted.internet import reactor
            self.paintTimer = reactor.callLater(PAINT_DELAY,
                                                self.paintCallback)

    def restack(self, window, new_above):
        index = self.windowstack.index(window)
        if index == len(self.windowstack) - 1:
            cur_above = None
        else:
            cur_above = self.windowstack[index + 1]
        if cur_above != new_above:
            self.windowstack.remove(window)
            if new_above is None:
                self.windowstack.append(window)
            else:
                index = self.windowstack.index(new_above)
                self.windowstack.insert(index, window)

    def paintCallback(self):
        self.paintTimer = None
        self.paint_all(self.allDamage)
        self.allDamage = None
        self.clipChanged = False


if __name__ == '__main__':
    import sys; log.startLogging(sys.stderr)
    # This adds lots of useful debugging output but is pretty slow.
    #defer.setDebugging(True)
    from twisted.internet import reactor
    d = defer.Deferred()
    d.addCallback(startup)
    @d.addErrback
    def die(fail):
        log.err(fail)
        reactor.stop()
    reactor.callWhenRunning(d.callback, reactor)
    reactor.callLater(60, reactor.stop)
    reactor.run()
