#!/usr/bin/env python

import os
import os.path as osp

from distutils import core, cmd, dep_util, log
from distutils.command import sdist, build


# XXX this assumes the toplevel source dir is $PWD

class BuildProto(cmd.Command):

    user_options = [
        ('force', 'f', 'forcibly build everything (ignore file timestamps)'),
        ]

    boolean_options = ['force']

    srcdir = 'xcb-proto'
    destdir = osp.join('tx11', 'proto')

    def initialize_options(self):
        self.force = None

    def finalize_options(self):
        self.set_undefined_options('build', ('force', 'force'))

    def run(self):
        names = [fn for fn in os.listdir(self.srcdir)
                 if osp.splitext(fn)[1] == '.xml']
        sources = [osp.join(self.srcdir, fn) for fn in names]
        targets = [osp.join(self.destdir, osp.splitext(fn)[0] + '.py')
                   for fn in names]
        if not sources:
            log.info('skipping proto rebuild (no sources)')
            return
        if not self.force:
            stale_sources, stale_targets = dep_util.newer_pairwise(sources,
                                                                   targets)
            if not stale_sources:
                log.info('skipping proto rebuild (up-to-date)')
                return
            # Rebuild everything, not just changed files, because one xml
            # file can refer to struct definitions in another
            # (we would have to build a full dep graph).
            # The pairwise comparisons above are technically bogus but
            # should suffice in practice.

        # Do these here so all this works with prebuilt proto files and no
        # tpfkax/py_client from dist tarballs.
        import py_client
        from tpfkax import state

        cache = state.XMLCache()
        for source in sources:
            module = py_client.PyModule(source, self.destdir, cache)
            module.register()
            module.resolve()
            module.generate()


class MySDist(sdist.sdist):

    """sdist extended to run build_proto first."""

    sub_commands = [('build_proto', None)]

    def run(self):
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)

        sdist.sdist.run(self)


class MyBuild(build.build):

    """build extended to run build_proto first."""

    sub_commands = [('build_proto', None)] + build.build.sub_commands


core.setup(name='Twisted-X11',
           version='0.1',
           description='Pure python implementation of the X11 protocol',
           author='Marien Zwart',
           author_email='tx11@emzet.cjb.net',
           packages=['tx11', 'tx11.proto', 'tx11.test'],
           # Copied from distutils/dist.py, with build_proto added.
           common_usage='''\
Common commands: (see '--help-commands' for more)

  setup.py build_proto   will (re)build the protocol files in place
  setup.py build         will build the package underneath 'build/'
  setup.py install       will install the package
''',
           cmdclass=dict(build_proto=BuildProto,
                         sdist=MySDist,
                         build=MyBuild),
           )
