
"""This module mainly contains the Namespace and Module class."""

from os.path import dirname, basename, join
from xml.etree.cElementTree import parse
from collections import defaultdict, namedtuple

from error import ResolveException


class XMLCache(object):

    def __init__(self):
        self.cache = {}

    def __call__(self, path):
        try:
            return self.cache[path]
        except KeyError:
            tree = parse(path)
            self.cache[path] = tree
            return tree


class Namespace(object):

    """Contains the naming information for an extension.

    Public fields:

    header is the header attribute ("header file" name).
    is_ext is true for extensions, false for xproto.
    major_version and minor_version are extension version info.
    ext_xname is the X extension name string.
    ext_name is the XCB extension name prefix.
    """

    def __init__(self, filename, parsefunc=parse):
        # Path info
        self.path = filename
        self.dir = dirname(filename)
        self.file = basename(filename)

        # Parse XML
        self.root = parsefunc(filename).getroot()
        self.header = self.root.get('header')
        self.ns = self.header + ':'

        # Get root element attributes
        if self.root.get('extension-xname', False):
            self.is_ext = True
            self.major_version = self.root.get('major-version')
            self.minor_version = self.root.get('minor-version')
            self.ext_xname = self.root.get('extension-xname')
            self.ext_name = self.root.get('extension-name')
            self.prefix = ('xcb', self.ext_name)
        else:
            self.is_ext = False
            self.ext_name = ''
            self.prefix = ('xcb',)


def _execute(module, namespace, parsefunc):
    for elt in namespace.root:
        name = elt.tag
        if name in ('request', 'enum', 'event', 'struct', 'error', 'union'):
            # Note there is more in the output dict than the tag names
            # we want to accept, so an explicit name check is desirable.
            typename = elt.get('name')
            qualname = namespace.prefix + (typename,)
            typeobj = module.output[name](qualname, elt)

            if name == 'request':
                module.add_type(namespace, typename, typeobj)
            elif name in ('event', 'error'):
                mapping = module.events if name == 'event' else module.errors
                typeobj.add_opcode(int(elt.get('number')), qualname)
                mapping[typename] = typeobj
                module.add_type(namespace, typename, typeobj)
            else:
                module.add_referenced_type(namespace, typename, typeobj)

        elif name in ('eventcopy', 'errorcopy'):
            typename = elt.get('name')
            mapping = module.events if name == 'eventcopy' else module.errors
            event = mapping[elt.get('ref')]
            event.add_opcode(int(elt.get('number')),
                             namespace.prefix + (typename,))
            mapping[typename] = event
            module.add_type(namespace, typename, event)

        elif name == 'import':
            if not module.has_import(elt.text):
                new_file = join(namespace.dir, '%s.xml' % (elt.text,))
                new_namespace = Namespace(new_file, parsefunc)
                _execute(module, new_namespace, parsefunc)
                module.add_import(elt.text, new_namespace)

        elif name == 'typedef':
            module.add_referenced_type(
                namespace, elt.get('newname'),
                module.get_type(elt.get('oldname')).item)

        elif name in ('xidtype', 'xidunion'):
            module.add_referenced_type(namespace, elt.get('name'),
                                       module.get_type('CARD32').item)

        else:
            raise ValueError('unsupported tag name %s' % (name,))


TypeTuple = namedtuple('TypeTuple', 'name item')


class Module(object):
    """This is the class that represents an entire XCB specification.

    Don't need to worry about this much except to declare it and to
    get the namespace.

    Public fields:
    namespace contains the namespace info for the spec.
    """

    def __init__(self, filename, output, parsefunc=parse):
        self.namespace = Namespace(filename, parsefunc)
        self.parsefunc = parsefunc
        self.output = output

        self.imports_by_name = {}
        self.imports_by_ext = {}
        self.types_by_qual = {}
        self.types_by_local = defaultdict(list)
        self.events = {}
        self.errors = {}
        self.all = []

        # Register some common types.
        cls = output['simple']
        for name, size in [('CARD8', 1),
                           ('CARD16', 2),
                           ('CARD32', 4),
                           ('INT8', 1),
                           ('INT16', 2),
                           ('INT32', 4),
                           ('BYTE', 1),
                           ('BOOL', 1),
                           ('char', 1),
                           ('float', 4),
                           ('double', 8),
                           ]:
            self.types_by_qual[name] = TypeTuple((name,), cls((name,), size))

        # XXX is this correct?
        self.types_by_qual['void'] = TypeTuple(('void',),
                                               cls(('void',), 1))

    def open(self):
        raise NotImplementedError('abstract method open not implemented')

    def close(self):
        raise NotImplementedError('abstract method close not implemented')

    def register(self):
        """This goes out and parses the rest of the XML."""
        _execute(self, self.namespace, self.parsefunc)

    def resolve(self):
        """Recursively resolve all types."""
        for (name, item) in self.all:
            if not item.resolved:
                item.resolve(self)
                item.resolved = True

    def generate(self):
        """Call all the output methods."""
        self.open()

        for (name, item) in self.all:
            item.out(self, name)

        self.close()

    def add_import(self, name, namespace):
        """Keeps track of what's been imported so far."""
        self.imports_by_name[name] = namespace
        self.imports_by_ext[namespace.ext_name] = namespace

    def has_import(self, name):
        """Check if something is already imported."""
        return name in self.imports_by_name

    def add_referenced_type(self, namespace, typename, item):
        """Keeps track of non-request/event/error datatypes."""
        key = namespace.ns + typename
        if key in self.types_by_qual:
            raise ValueError('duplicate key %s' % (key,))
        name = namespace.prefix + (typename,)
        val = TypeTuple(name, item)
        self.types_by_qual[key] = val
        self.types_by_local[typename].append(val)
        self.add_type(namespace, typename, item)

    def get_type(self, name):
        """Return a tuple (name, typeobj) for an optionally qualified name."""
        result = self.types_by_qual.get(name)
        if result is not None:
            return result

        result = self.types_by_qual.get(self.namespace.ns + name)
        if result is not None:
            return result

        results = self.types_by_local.get(name)
        if results is not None:
            if len(results) > 1:
                raise ResolveException('ambiguous key %s' % (name,))
            return results[0]

        raise ResolveException('Type %s not found' % (name,))

    def add_type(self, namespace, typename, item):
        """Keeps track of request/event/error datatypes."""
        if namespace is self.namespace:
            self.all.append((namespace.prefix + (typename,), item))
