
"""This module contains the classes which represent XCB data types."""

from expr import Field, Expression


_placeholder_byte = object()


class Type(object):

    """Abstract base class for all XCB data types.

    Contains default fields, and some abstract methods.
    """

    def __init__(self, name):

        """Default structure initializer.  Sets up default fields.

        Public fields:
        name is a tuple of strings specifying the full type name.
        size is the size of the datatype in bytes, or None if variable-sized.
        booleans for identifying subclasses, because I can't figure out isinstance().
        """
        self.name = name
        self.size = None
        self.resolved = False

    def resolve(self, module):
        """Abstract method for resolving a type.

        This should make sure any referenced types are already declared.
        """
        raise Exception('abstract resolve method not overridden!')

    def out(self, module, name):
        """Abstract method for outputting code.

        These are declared in the language-specific subclasses.
        """
        raise Exception('abstract out method not overridden!')

    def make_member_of(self, module, complex_type, field_type, field_name, visible, wire, auto):
        """Default method for making a data type a member of a structure.

        Extend this if the data type needs to add an additional length field or something.

        module is the global module object.
        complex_type is the structure object.
        see Field for the meaning of the other parameters.
        """
        new_field = Field(self, field_type, field_name, visible, wire, auto)

        # We dump the _placeholder_byte if any fields are added.
        for (idx, field) in enumerate(complex_type.fields):
            if field is _placeholder_byte:
                assert new_field.type.size == 1
                complex_type.fields[idx] = new_field
                return

        complex_type.fields.append(new_field)

class SimpleType(Type):

    """Derived class which represents a cardinal type like CARD32 or char.

    Any type which is typedef'ed to cardinal will be one of these.

    Public fields added:
    none
    """

    def __init__(self, name, size):
        Type.__init__(self, name)
        self.size = size
        self.resolved = True


class Enum(SimpleType):

    """Derived class which represents an enum.  Fixed-size.

    Public fields added:
    values contains a list of (name, value) tuples.  value is empty, or a number.
    bits contains a list of (name, bitnum) tuples.  items only appear if specified as a bit. bitnum is a number.
    """

    def __init__(self, name, elt):
        SimpleType.__init__(self, name, 4)
        self.values = []
        self.bits = []
        for item in elt:
            # First check if we're using a default value
            if not len(item):
                self.values.append((item.get('name'), ''))
                continue

            # An explicit value or bit was specified.
            value = item[0]
            tag = value.tag
            text = value.text
            if tag == 'value':
                self.values.append((item.get('name'), text))
            elif value.tag == 'bit':
                self.values.append((item.get('name'), str(1 << int(text))))
                self.bits.append((item.get('name'), text))
            else:
                raise Exception('unreachable?')


class ListType(Type):

    """Represents a fixed- or variable-sized list of some other datatype.

    Public fields added:
    member is the datatype of the list elements.
    parent is the structure type containing the list.
    expr is an Expression object containing the length information.
    """

    def __init__(self, elt, member, parent):
        Type.__init__(self, member.name)
        self.member = member
        self.parent = parent

        if elt.tag == 'list':
            self.expr = Expression(elt[0] if len(elt) else elt, self)
        elif elt.tag == 'valueparam':
            self.expr = Expression(elt, self)

        if member.size is None or self.expr.value is None:
            self.size = None
        else:
            self.size = member.size * self.expr.value

    def make_member_of(self, module, complex_type, field_type, field_name, visible, wire, auto):
        if self.size is None:
            # We need a length field.
            # Ask our Expression object for it's name, type, and whether it's on the wire.
            lenfid = self.expr.lenfield_type
            lenfield_name = self.expr.lenfield_name
            lenwire = self.expr.lenwire
            needlen = True

            # See if the length field is already in the structure.
            for field in self.parent.fields:
                if field.field_name == lenfield_name:
                    needlen = False

            # It isn't, so we need to add it to the structure ourself.
            if needlen:
                lenfield_type, type = module.get_type(lenfid)
                type.make_member_of(module, complex_type, lenfield_type, lenfield_name, True, lenwire, False)

        # Add ourself to the structure by calling our original method.
        Type.make_member_of(self, module, complex_type, field_type, field_name, visible, wire, auto)

    def resolve(self, module):
        if not self.member.resolved:
            self.member.resolve(module)
            self.member.resolved = True

        # Find my length field again.  We need the actual Field object in the expr.
        # This is needed because we might have added it ourself above.
        if self.size is None:
            for field in self.parent.fields:
                if field.field_name == self.expr.lenfield_name and field.wire:
                    self.expr.lenfield = field
                    break


class ExprType(Type):

    """Represents a fixed-size exprfield.

    Public fields added:
    expr is an Expression object containing the value of the field.
    """

    def __init__(self, elt, member, parent):
        Type.__init__(self, member.name)
        self.member = member
        self.parent = parent

        self.expr = Expression(elt[0], self)

        self.size = member.size

    def resolve(self, module):
        if not self.member.resolved:
            self.member.resolve(module)
            self.member.resolved = True


class SizedPadType(Type):

    """Represents a padding field of given size."""

    def __init__(self, size):
        # XXX this does not really have a name, or does it?
        Type.__init__(self, None)
        self.size = size
        self.resolved = True


class ComplexType(Type):

    """Base type for all structure types (sequence of variable-typed fields).

    Public fields added:
    fields is an array of Field objects describing the structure fields.
    """

    def __init__(self, name, elt):
        Type.__init__(self, name)
        self.elt = elt
        self.fields = []
        self.size = -1

    def resolve(self, module):
        pads = 0

        # Resolve all of our field datatypes.
        for child in self.elt:
            value_mask_pad = None
            if child.tag == 'pad':
                field_name = 'pad' + str(pads)
                # XXX is this sane?
                field_type = None
                type = module.output['pad'](int(child.get('bytes')))
                pads += 1
                visible = False
            elif child.tag == 'field':
                field_name = child.get('name')
                field_type, type = module.get_type(child.get('type'))
                visible = True
            elif child.tag == 'exprfield':
                field_name = child.get('name')
                field_type, item = module.get_type(child.get('type'))
                type = module.output['expr'](child, item, self)
                visible = False
            elif child.tag == 'list':
                field_name = child.get('name')
                field_type, item = module.get_type(child.get('type'))
                type = module.output['list'](child, item, self)
                visible = True
            elif child.tag == 'valueparam':
                field_name = child.get('value-list-name')
                field_type, item = module.get_type('CARD32')
                type = module.output['list'](child, item, self)
                visible = True
                value_mask_pad = child.get('value-mask-pad')
            else:
                # Hit this on Reply
                if child.tag != 'reply':
                    raise ValueError('unsupported tag %s' % (child.tag,))
                continue

            # Add the field to ourself
            type.make_member_of(module, self, field_type, field_name, visible, True, False)
            # Recursively resolve the type (could be another structure, list)
            if not type.resolved:
                type.resolve(module)
                type.resolved = True

            # Add a value-mask-pad if necessary
            if value_mask_pad is not None:
                vmp_field_name = 'pad' + str(pads)
                pads += 1
                vmp_type = module.output['pad'](int(value_mask_pad))
                # XXX is this sane?
                vmp_field_type = None
                vmp_type.make_member_of(module, self, vmp_field_type, vmp_field_name, False, True, False)
                if not vmp_type.resolved:
                    vmp_type.resolve(module)
                    vmp_type.resolved = True

        # Replace the _placeholder_byte with padding if it's still there, and
        # figure out how big we are
        self.size = 0
        for i, m in enumerate(self.fields):
            if m is _placeholder_byte:
                padtype = module.output['pad'](1)
                self.fields[i] = m = Field(padtype, padtype.name,
                                           'placeholder', False, True, False)
            if not m.wire:
                continue
            if m.type.size is None:
                self.size = None
                break
            self.size = self.size + m.type.size


class Struct(ComplexType):
    """Represents a struct data type."""


class Union(ComplexType):

    """Represents a union data type."""

    def resolve(self, module):
        ComplexType.resolve(self, module)
        # Override the size the superclass just calculated.
        # The size of a union is the size of the largest field.
        # If any of the fields is unsized so is a struct or union,
        # so for that case the superclass gets it right:
        if self.size is not None:
            self.size = 0
            for m in self.fields:
                if m.wire and m.type.size > self.size:
                    self.size = m.type.size


class Reply(ComplexType):

    """Represents a reply.  Only found as a field of Request."""

    def resolve(self, module):
        # Add the automatic protocol fields
        tcard8 = module.get_type('CARD8').item
        tcard16 = module.get_type('CARD16').item
        tcard32 = module.get_type('CARD32').item
        self.fields.append(Field(tcard8, tcard8.name, 'response_type', False, True, True))
        self.fields.append(_placeholder_byte)
        self.fields.append(Field(tcard16, tcard16.name, 'sequence', False, True, True))
        self.fields.append(Field(tcard32, tcard32.name, 'length', False, True, True))
        ComplexType.resolve(self, module)


class Request(ComplexType):

    """Represents a request.

    Public fields added:
    reply contains the reply datatype or None for void requests.
    opcode contains the request number.
    """

    def __init__(self, name, elt):
        ComplexType.__init__(self, name, elt)
        self.reply = None
        self.opcode = elt.get('opcode')

        for child in elt:
            if child.tag == 'reply':
                self.reply = Reply(name, child)

    def resolve(self, module):
        # Add the automatic protocol fields
        tcard8 = module.get_type('CARD8').item
        tcard16 = module.get_type('CARD16').item
        if module.namespace.is_ext:
            self.fields.append(Field(tcard8, tcard8.name, 'major_opcode', False, True, True))
            self.fields.append(Field(tcard8, tcard8.name, 'minor_opcode', False, True, True))
            self.fields.append(Field(tcard16, tcard16.name, 'length', False, True, True))
        else:
            self.fields.append(Field(tcard8, tcard8.name, 'major_opcode', False, True, True))
            self.fields.append(_placeholder_byte)
            self.fields.append(Field(tcard16, tcard16.name, 'length', False, True, True))

        ComplexType.resolve(self, module)

        if self.reply:
            # There should be no way for our reply to get resolved
            # except through us:
            assert not self.reply.resolved
            self.reply.resolve(module)
            self.reply.resolved = True


class ComplexWithOpcodes(ComplexType):

    """Represents a type with one or more opcodes.

    Public fields added:
    opcodes is a dictionary of name -> opcode number, for copies.
    """

    def __init__(self, name, elt):
        ComplexType.__init__(self, name, elt)
        self.opcodes = {}

    def add_opcode(self, opcode, name):
        self.opcodes[name] = opcode


class Event(ComplexWithOpcodes):

    """Represents an event data type.

    Public fields added:
    opcodes is a dictionary of name -> opcode number, for eventcopies.
    """

    def __init__(self, name, elt):
        ComplexWithOpcodes.__init__(self, name, elt)

        tmp = elt.get('no-sequence-number')
        self.has_seq = (tmp is None or tmp.lower() == 'false' or tmp == '0')

    def resolve(self, module):
        # Add the automatic protocol fields
        tcard8 = module.get_type('CARD8').item
        tcard16 = module.get_type('CARD16').item
        self.fields.append(Field(tcard8, tcard8.name, 'response_type', False, True, True))
        if self.has_seq:
            self.fields.append(_placeholder_byte)
            self.fields.append(Field(tcard16, tcard16.name, 'sequence', False, True, True))
        ComplexType.resolve(self, module)


class Error(ComplexWithOpcodes):

    """Represents an error data type."""

    def resolve(self, module):
        # Add the automatic protocol fields
        tcard8 = module.get_type('CARD8').item
        tcard16 = module.get_type('CARD16').item
        self.fields.append(Field(tcard8, tcard8.name, 'response_type', False, True, True))
        self.fields.append(Field(tcard8, tcard8.name, 'error_code', False, True, True))
        self.fields.append(Field(tcard16, tcard16.name, 'sequence', False, True, True))
        ComplexType.resolve(self, module)
