
"""Reimplementation of (parts of) libXau.

Writing XAuth data is unsupported for now.

This just does libXau (Xauthority) auth. Xdmcp is unsupported.
"""

from __future__ import absolute_import, with_statement

import struct
from socket import gethostname
from os import environ, path as osp

from twisted.internet import address

from . import etx11


# Only the libXau constants we actually support...
FAMILY_LOCAL = 256

# Schemes to try in order of preference
# (used as default kwarg, so keep this immutable).
DEFAULT_SCHEMES = ('MIT-MAGIC-COOKIE-1',)


def _getFilename():
    """Figure out where Xauthority data lives.

    Uses the env vars XAUTHORITY or HOME. Returns None if both are unset.
    """
    # Directly based on XauFileName.
    path = environ.get('XAUTHORITY')
    # Treat the empty string the same as unset here.
    if path:
        return path

    path = environ.get('HOME')
    if path is not None:
        return osp.join(path, '.Xauthority')

    # Win32 code omitted.
    return None


# This bit is based on both libXau and the xpyb wiki,
# with some xcb_auth.c thrown in for good measure.
# In other words: this is a mess in need of testing.

def _read_counted_string(f):
    return f.read(struct.unpack('!H', f.read(2))[0])

def getXAuthData():
    """Parse the XAUTHORITY file if there is one."""
    # Not a defaultdict to avoid confusing calling code.
    result = {}
    filename = _getFilename()
    if filename is None:
        return result
    try:
        with open(filename, 'rb') as xauthfile:
            while True:
                data = xauthfile.read(2)
                if not data:
                    # Done!
                    return result

                family, = struct.unpack('!H', data)
                addr = _read_counted_string(xauthfile)
                number = _read_counted_string(xauthfile)
                name = _read_counted_string(xauthfile)
                data = _read_counted_string(xauthfile)

                key = (family, addr, number)
                methods = result.setdefault(key, {})

                if name in methods:
                    raise etx11.XauthorityException(
                        'duplicate method %s for %r' % (name, key))
                methods[name] = data

    except struct.error:
        raise etx11.XauthorityException('"%s" truncated or corrupt?' %
                                        (filename,))
    except EnvironmentError, e:
        raise etx11.XauthorityException('error reading "%s": %s' %
                                        (filename, e))


def getByAddress(peer, display, schemes=DEFAULT_SCHEMES):
    """Try to get a promising auth method and data from Xauthority.

    address is the IAddress and display is the number of the display
    we are connected to.

    schemes is a sequence of schemes to try (preferred one first).

    Returns a (scheme, data) tuple or None.
    """
    if isinstance(peer, address.UNIXAddress):
        # This just has a path, which is useless to us...
        hostname = gethostname()
        kind = FAMILY_LOCAL
    elif isinstance(peer, address.IPV4Address):
        hostname = peer.host
        # TODO: if this is a dotted quad Xauth has it as 4 bytes
        # (so Xauth has '\xc0\xa8\x00\x02' where I think I will get
        # '192.168.0.2' from Twisted). Need to convert those.
        # TODO: what's the right kind value for these?
        kind = None
    else:
        # Uhoh, unsupported address scheme...
        return None

    return getByName(kind, hostname, str(display), schemes)


def getByName(kind, hostname, display, schemes=DEFAULT_SCHEMES):
    """Like getByAddress but takes strings for address and display.

    kind is a FAMILY_* value.
    """
    xauth = getXAuthData()

    # Intended effect here is that we prefer the best scheme, and
    # prefer the most specific key (proto, hostname, etc) after that.
    d = {}
    d.update(xauth.get((0, hostname, display), {}))
    d.update(xauth.get((kind, hostname, display), {}))
    for scheme in schemes:
        data = d.get(scheme)
        if data is not None:
            return scheme, data

    return None
