
"""Core protocol implementation.

This module's "connect" function is the starting point for users of
twisted-x11.
"""

from __future__ import absolute_import, with_statement

import sys
import struct
from collections import deque
from os import environ

from twisted.python import failure, log
from twisted.internet import protocol, defer
from twisted.protocols import stateful

from . import etx11, base, auth
from .proto import xproto, bigreq, xc_misc


MAJOR_VERSION = 11
MINOR_VERSION = 0
TCP_PORT = 6000
UNIX_BASE = '/tmp/.X11-unix/X'

KEYMAP_NOTIFY_EVENT = 11

ERROR = 0
REPLY = 1
XGE_EVENT = 35

if sys.byteorder == 'big':
    BOM = ord('B')
else:
    BOM = ord('l')

_struct_BBH_pack = struct.Struct('BBH').pack
_struct_BB2xI_pack = struct.Struct('BB2xI').pack
_struct_BBHI_from = struct.Struct('BBHI').unpack_from


def _pad(s):
    return s + '\0' * (-len(s) & 3)


# TODO: connect to a specific fd?
# TODO: support decnet (but probably not).

def _parseDisplay(name=None, screen=0):
    """Parse a DISPLAY string passed in or grabbed from the env.

    screen is the preferred screen if not set in the display string.
    """
    if name is None:
        name = environ['DISPLAY']

    if not name:
        raise ValueError('no display set')

    # Protocol being empty is fine.
    proto, sep, name = name.rpartition('/')
    # Handle proto unset differently from proto empty.
    # XXX not sure how useful this actually is.
    if not sep:
        proto = None

    host, sep, name = name.rpartition(':')
    if not sep:
        raise ValueError('invalid display (":" not found)')

    display, sep, screenstring = name.partition('.')
    try:
        display = int(display)
    except ValueError:
        raise ValueError('invalid display (display not a number)')

    if sep:
        try:
            screen = int(screenstring)
        except ValueError:
            raise ValueError('invalid display (screen not a number)')

    return host, proto, display, screen


def ignoreConnectionLost(reason=None):
    """Dummy connectionLost handler."""


def _noReply(buf):
    raise etx11.ProtocolException('reply for void request')


class UnknownError(etx11.RemoteError):
    """Raised when an unknown error (unregistered extension?) is detected."""
    def __init__(self, buf):
        etx11.RemoteError.__init__(self)
        self.error_code, = struct.unpack_from('xB2x4x', buf)
        self.buf = buf


class UnknownEvent(base.Event):
    def __init__(self, buf):
        base.Event.__init__(self)
        self.response_type, = struct.unpack_from('B', buf)
        self.buf = buf


class X11Protocol(stateful.StatefulProtocol):

    def __init__(self, display, xauth=None,
                 extensions=(), requiredExtensions=None,
                 shutdownHandler=ignoreConnectionLost,
                 eventHandler=lambda conn, event: None):
        """Initialize.

        display is the number of the display we are connecting to.
        xauth is a tuple (scheme, data) or None.
        extensions is a sequence of Extension objects to request.
        requiredExtensions is a sequence of required Extension objects.
        Setting this to None means use the "extensions" value.
        shutdownHandler is just like connectionLost.
        eventHandler is called with self and an Event object on events.
        This is assigned to self.eventHandler, simply reassigning that
        is supported.
        See the event module for interesting values for this.

        if xauth is None the value passed as display, the address we
        connect to and the Xauthority file are used (see the auth module).
        """
        self.auth = xauth
        self.display = display
        self.eventHandler = eventHandler
        # Notice this does the right thing if an implicitly used
        # extension (xc_misc, bigreq) is also requested explicitly.
        if requiredExtensions is None:
            requiredExtensions = extensions
        extensions = set(extensions)
        # We may use xc_misc if available for xid updates.
        extensions.add(xc_misc.XCMiscExtension)
        # BigRequests needs special handling to actually enable it asap.
        # Make sure it is first in the list (code below relies on this).
        extensions.discard(bigreq.BigRequestsExtension)
        extensions = [bigreq.BigRequestsExtension] + list(extensions)

        # Called back with setup data once we have it.
        # Erred back if we fail before receiving that data.
        # Careful: do not drop this on errors, or callers cannot get
        # at the failure if we fail synchronously during startup...
        self._initialDeferred = defer.Deferred()

        @self._initialDeferred.addCallback
        def setupReceived(pair):
            status, setup = pair
            if status == 0:
                # Failure.
                raise etx11.ConnectionFailed(xproto.SetupFailed(setup))
            if status == 1:
                # Success.
                self.setup = xproto.Setup(setup)
                return self.setup
            elif status == 2:
                # Authenticate.
                raise etx11.ConnectionDenied(xproto.SetupAuthenticate(setup))
            else:
                # huh?
                raise etx11.ProtocolException('weird status %s' % (status,))

        @self._initialDeferred.addCallback
        def _initialSetup(setup):
            # The mask is a "single contiguous set of bits (at least 18)".
            # We own the resources with some of those bits set and all
            # of the bits in "base" set.
            # On xorg-server-1.5.3 the mask includes the least significant
            # bit, so we simply own the range base <= xid <= base + mask.
            # But that is *not* required: if I skim the protocol docs
            # correctly base could have less significant bits set than mask.

            # This picks just the least significant bit in mask.
            self._xidInc = setup.resource_id_mask & -setup.resource_id_mask
            # _xidLast is the last id we handed out.
            # Start that at one increment before the lowest id:
            self._xidLast = setup.resource_id_base - self._xidInc
            # The highest id that is still valid.
            self._xidMax = setup.resource_id_base | setup.resource_id_mask
            # List of deferreds waiting for new ids:
            self._xidWaiting = []

            self._replyqueue = deque()
            self._maxNonVoid = 0
            self._replyseq = 1

            self.core = xproto.Core(self)

            self._events = self.core._events.copy()
            self._errors = self.core._errors.copy()

            self.maximumRequestLength = setup.maximum_request_length

            # The idea here is that we want to enable bigreq before we do
            # anything involving nontrivial requests (or we have to do it
            # asynchronously in sendRequest, which would mean either
            # returning a deferred there or having to queue up further
            # sendRequest calls internally while waiting for the reply).
            # But because other QueryExtension calls had better not need
            # bigreq we can run those in parallel.
            ds = [self.registerExtension(ext) for ext in extensions]
            @defer.inlineCallbacks
            def maybeEnableBigreq(present):
                if present:
                    reply = yield self.bigreq.enable()
                    self.maximumRequestLength = reply.maximum_request_length
                defer.returnValue(present)
            ds[0].addCallback(maybeEnableBigreq)
            return defer.gatherResults(ds)

        if requiredExtensions:
            @self._initialDeferred.addCallback
            def detectMissing(extensionresults):
                missing = set(requiredExtensions).intersection(
                    ext for ext, present in zip(extensions, extensionresults)
                    if not present)
                if missing:
                    raise etx11.ExtensionMissing(missing)

        # If we manage to lose our connection *during* init we fail
        # (errback) our initfunc. If we lose our connection after
        # successful initialization we pass it on to shutdownhandler.
        # If we lose our connection after *failing* init we need to
        # ignore that to avoid duplicate failures in us or our caller.

        def _cb(ignored):
            self.connectionLost = shutdownHandler
            return self
        def _eb(passthrough):
            self.connectionLost = ignoreConnectionLost
            self.transport.loseConnection()
            return passthrough
        self._initialDeferred.addCallbacks(_cb, _eb)

    def getInitialState(self):
        return self._setupHeadReceived, 8

    def connectionMade(self):
        try:
            if self.auth is None:
                xauth = auth.getByAddress(self.transport.getPeer(),
                                          self.display)
                if xauth is None:
                    xauth = ('', '')
            else:
                xauth = self.auth

            authname, authinfo = xauth

            self.transport.writeSequence([
                    _pad(struct.pack(
                            'BxHHHH', BOM, MAJOR_VERSION, MINOR_VERSION,
                            len(authname), len(authinfo))),
                    _pad(authname),
                    _pad(authinfo)])
        except Exception:
            self._initialDeferred.errback()
            self.transport.loseConnection()

    def connectionLost(self, reason):
        # This function gets overwritten during initialization.
        # We lost the connection before we finished initializing, so fail.
        self._initialDeferred.errback(reason)

    def generateId(self):
        """Generate a new id.

        This method returns a deferred that has almost always already
        been called back. Make sure the case where the deferred is
        called back asynchronously is handled, but feel free to treat
        method as synchronous as far as performance is concerned.

        For the curious: we have a pool of ids to hand out. If the
        pool empties we need to request new ids, which means a server
        roundtrip. That only happens for long-running clients, and even
        then it will hopefully not happen frequently.

        How frequently this happens for long-running clients is
        UNTESTED and probably depends on how they use their ids.
        """
        # Thought: we could try to be smarter here and request fresh ids
        # *before* we actually run out. But depending on how the server
        # implements it that could perform horribly if we hit it too often.
        # For example GetXIDList in at least some xorg-server versions
        # does a linear search starting at our resource base.
        # So you really do not want to call that repeatedly while using
        # many (low) ids... So just do the obvious thing here and
        # revisit this once we actually hit that code.
        last = self._xidLast
        if last != self._xidMax:
            # The common case: we still have ids to hand out.
            last = last + self._xidInc
            self._xidLast = last
            return defer.succeed(last)

        # We ran out.
        if self.xc_misc is None:
            return defer.fail(etx11.NoMoreIds())

        # (paranoia: add to _xidWaiting before requesting more ids in case
        # we run against something responding to the request synchronously)
        d = defer.Deferred()
        d.addCallback(lambda ignored: self.generateId())
        self._xidWaiting.append(d)
        if len(self._xidWaiting) == 1:
            def gotRange(reply):
                # This one is different from the core protocol.
                # We get a start id and a length. I assume that means
                # we should now start using an increment of 1.
                self._xidInc = 1
                self._xidLast = reply.start_id - 1
                self._xidMax = self._xidLast + reply.count
                # Careful: the callbacks we trigger here reenter...
                xidWaiting = self._xidWaiting
                self._xidWaiting = []
                if not reply.count:
                    # The server gave us no ids! Fail our pending calls
                    # or we could hit an infinite loop here:
                    for d in xidWaiting:
                        d.errback(etx11.NoMoreIds())
                else:
                    for d in xidWaiting:
                        d.callback(None)
            def failedRange(f):
                # Again be careful about reentrance.
                xidWaiting = self._xidWaiting
                self._xidWaiting = []
                for d in xidWaiting:
                    d.errback(f)
            self.xc_misc.getXIDRange().addCallbacks(gotRange, failedRange)

        return d

    def _setupHeadReceived(self, data):
        """Initial state: determine which setup reply to receive."""
        status, length = struct.unpack('B5xH', data)

        def setupReceived(moredata):
            self._initialDeferred.callback((status, data + moredata))
            return self._packetHeadReceived, 32

        return setupReceived, 4 * length

    @defer.inlineCallbacks
    def registerExtension(self, extension):
        """Call with an Extension subclass. Returns a deferred.

        Sets self.theextension to None and returns False if it is not
        present on the server.

        Sets self.theextension to an instance and returns True if it is.
        """
        reply = yield self.core.queryExtension(extension._name)

        if not reply.present:
            setattr(self, extension._pyname, None)
            defer.returnValue(False)

        setattr(self, extension._pyname, extension(self, reply.major_opcode))

        # XXX we may actually have to support overlap below.
        # If newer versions of extensions add events and we support a newer
        # version than the server we should let the ones with the lower
        # relative (to first_blah) number win.
        # But we can obviously only detect that if we actually register
        # both extensions. So I'm guessing this is avoided in practice.

        for mapping, target, first in [
            (extension._events, self._events, reply.first_event),
            (extension._errors, self._errors, reply.first_error)]:
            if mapping and not first:
                raise etx11.ProtocolException('incomplete extension')
            for n, klass in mapping.iteritems():
                if first + n in target:
                    raise etx11.ProtocolException(
                        'overlapping event/error index (%s, %s)' % (
                            klass, target[first + n]))
                target[first + n] = klass

        defer.returnValue(True)

    def _packetHeadReceived(self, data):
        """Read the first 32 bytes, decide how much more to read.

        We alternate between this and the _packetReceived state.
        """
        response_type, error_code, seqnr, length = _struct_BBHI_from(data)
        self._packet_head = (response_type, error_code, seqnr, length, data)
        # Only replies and XGE events have more data
        # (for regular events and errors the field means something else).
        # XXX the XGE stuff is untested! Need a newer X server for that...
        if response_type in (REPLY, XGE_EVENT):
            extra = length * 4
        else:
            extra = 0
        return self._packetReceived, extra

    def _packetReceived(self, data):
        """Read everything but the first 32 bytes and handle the packet.

        We alternate between this and the _packetHeadReceived state.
        """
        response_type, error_code, seqnr, length, head = self._packet_head
        data = head + data
        #print seqnr, response_type, len(data), repr(data)

        try:
            # Every request we sent before this one has apparently
            # succeeded, so drop their deferreds.
            try:
                while self._replyseq & 0xffff != seqnr:
                    d, replyclass = self._replyqueue.popleft()
                    # Complain if we expected a reply but never received it:
                    if not (replyclass is _noReply or d.called):
                        raise etx11.ProtocolException('missing reply')
                    self._replyseq += 1
                if not self._replyqueue:
                    raise IndexError()
            except IndexError:
                raise etx11.ProtocolException('reply queue drained')

            # We intentionally do not pop from the queue for replies and
            # errors. Events after this one use this sequence number, so
            # it is simpler to keep this one queued. If we get a second
            # reply or error for the same seqnr we simply deal with the
            # AlreadyCalledError.
            if response_type == REPLY:
                d, replyclass = self._replyqueue[0]
                reply = replyclass(data, length)
                try:
                    d.callback(reply)
                except defer.AlreadyCalledError:
                    raise etx11.ProtocolException('duplicate reply')
            elif response_type == ERROR:
                d, replyclass = self._replyqueue[0]
                # I think unknown error codes should not be fatal here,
                # since if someone manages to send an extension request
                # without registering the extension we will hit them,
                # but we can still continue safely.
                errorclass = self._errors.get(error_code, UnknownError)
                try:
                    d.errback(failure.Failure(errorclass(data)))
                except defer.AlreadyCalledError:
                    raise etx11.ProtocolException('duplicate error')
            else:
                # Rumour has it that the most significant bit is set for
                # events from SendEvent (client sent events).
                # Like unknown errors unknown events do not have to be fatal.
                eventclass = self._events.get(response_type & ~0x80,
                                              UnknownEvent)
                event = eventclass(data)
                event.client_sent = bool(response_type & 0x80)
                self.eventHandler(self, event)
        except Exception, e:
            # XXX now what? If we just reraise we will get really
            # confused: we need to go back to _packetHeadReceived
            # state to have any chance of recovering. For now just
            # panic and drop the connection, but perhaps higher level
            # code should be allowed to continue here...
            log.err()
            self.transport.loseConnection()

        return self._packetHeadReceived, 32

    def sendRequest(self, major, minor, data, replyclass=_noReply):
        """Send a request to the server. Normally only called internally.

        major and minor are numbers, 0 <= n <= 255.

        data is a string.

        replyclass is the class used to parse the reply or omitted for void.
        """
        # XXX this will do for now but I am not happy about this.
        #
        # bigrequests turn out to be annoying: we need to effectively
        # shift the first two bytes of the request to the left to make
        # room. If we have decent writable buffers we can deal with
        # that by taking a buffer with 4 bytes of slack at the start.
        # But if we do that we must make sure we introduce no extra
        # copies accidentally. This approach is a bit simpler and
        # harder to mess up.

        # Could pass the quadlength in directly,
        # but that's pushing micro-optimization too far.
        assert not len(data) % 4, 'need padding'
        quadlength = (len(data) >> 2) + 1

        if quadlength <= self.setup.maximum_request_length:
            # We don't need BIGREQUESTS.
            head = _struct_BBH_pack(major, minor, quadlength)
        elif quadlength <= self.maximumRequestLength:
            # Issue a "big" request.
            head = _struct_BB2xI_pack(major, minor, quadlength + 1)
        else:
            raise etx11.ProtocolException('request too long')

        # This does not work around the bug described in glx.xml with
        # GetFBConfigs.

        # We must not end up with a run of queued void requests as
        # long as we have unique 16-bit tails of sequence numbers, or
        # we could end up attributing a server message to the wrong
        # request. Note that we can keep emitting void requests just
        # fine as long as we know they get handled. Throw in a dummy
        # request if we are about to queue up too many voids.

        # (some numbers: moving around a window in the compositor demo
        # makes us emit thousands of requests per second, all void.
        # The queue size is pretty stable at around 100.)

        # (this recurses, make sure that is safe!)

        if replyclass is _noReply:
            # Two cases to consider:
            # - if maxNonVoid < replyseq everything queued is void.
            #   So total queue length must not exceed our max.
            # - if maxNonVoid >= replyseq we can subtract everything up
            #   to and including the maxNonVoid request.
            # The math here is off a bit, but that's ok, we just fire
            # the sync request a little early.
            # HACK: check if the queue length itself is large enough first,
            # and skip the full check if it isn't. It almost always isn't.
            if len(self._replyqueue) >= 0xfffa and len(self._replyqueue) - \
                    max(0, self._maxNonVoid - self._replyseq) >= 0xfffa:
                # XXX TODO what if this fails?
                self.core.getInputFocus()
        else:
            self._maxNonVoid = self._replyseq + len(self._replyqueue)

        #print 'OUT', repr(data.getvalue())

        self.transport.writeSequence([head, data])

        d = defer.Deferred()
        self._replyqueue.append((d, replyclass))
        return d


def connect(reactor, display=None, screen=-1, **kwargs):
    """Connect to a server and initialize.

    reactor is the reactor to connect with.

    display is a string. os.environ['DISPLAY'] is used if omitted.

    screen is the preferred screen index, overridden by DISPLAY if set
    there. This is set as conn.preferredScreen. Note this is
    determined from the DISPLAY variable used or the explicit fallback
    passed in, not handed to us by the server.

    Other keyword arguments are passed to X11Protocol.
    """
    host, proto, display, screen = _parseDisplay(display, screen)

    clientCreator = protocol.ClientCreator(reactor, X11Protocol, display,
                                           **kwargs)

    if proto or (host and host != 'unix'):
        if proto == 'tcp':
            d = clientCreator.connectTCP(host, TCP_PORT + display)
        raise ValueError('unsupported protocol %s' % (proto,))
    else:
        d = clientCreator.connectUNIX(UNIX_BASE + str(display))

    @d.addCallback
    def init(conn):
        conn.preferredScreen = screen
        return conn._initialDeferred

    return d
