
"""Event handler implementations for use with X11Protocol.

These are still in flux.
"""


from __future__ import absolute_import, with_statement

from collections import defaultdict
from itertools import chain

from twisted.python import log
from twisted.internet import defer


class SimpleHandler(object):

    def __init__(self):
        self.clear()

    def clear(self):
        self.eventhandlers = defaultdict(list)
        self.globalhandlers = []

    def __call__(self, conn, event):
        for handler in chain(self.globalhandlers,
                             self.eventhandlers[type(event)]):
            # Trap errors from each handler, so one handler failing
            # does not cause the others to fail.
            # TODO maybe offer a hook here, allowing the app to panic?
            try:
                handler(conn, event)
            except:
                log.err()

    def addEventHandler(self, handler, *classes):
        """Register an event handler for some or all event types.

        The handler is called with this connection and the event as arguments.

        Other arguments are event classes the handler is interested
        in. Passing no classes is equivalent to passing all of them.

        Registering the same handler more than once currently means it
        will get called more than once.
        """
        if classes:
            for klass in classes:
                self.eventhandlers[klass].append(handler)
        else:
            self.globalhandlers.append(handler)

    def removeEventHandler(self, handler, *classes):
        """Reverse addEventHandler.

        Calling this with the same arguments passed to addEventHandler
        (and the same number of times) is recommended, although
        passing classes in a different order or number of calls
        currently also works.
        """
        if classes:
            for klass in classes:
                self.eventhandlers[klass].remove(handler)
        else:
            self.globalhandlers.remove(handler)


# All tests above should apply to the SynchronousHandler too...
class SynchronousHandler(SimpleHandler):

    """You probably do not want to use this.

    This is intended as a porting aid. It allows event handlers to
    return a deferred and queues further events until that deferred
    fires.

    It uses a DeferredLock to do this, which is exposed as self.lock.
    Code not triggered from an event can acquire the same lock to
    synchronize.
    """

    # Notice this relies on DeferredLock acting as a fifo. We have
    # some basic tests that will hopefully fail if it stops doing that.
    # (they fail if it turns into a lifo but not if it gets weirder).

    def clear(self):
        SimpleHandler.clear(self)
        self.lock = defer.DeferredLock()

    def __call__(self, conn, event):
        for handler in chain(self.globalhandlers,
                             self.eventhandlers[type(event)]):
            self.lock.run(handler, conn, event)
