
from twisted.trial import unittest
from twisted.internet import defer

from tx11 import event, base


class DummyEvent(base.Event):
    pass

class AnotherDummyEvent(base.Event):
    pass


class SimpleHandlerTest(unittest.TestCase):

    def setUp(self):
        self.handler = event.SimpleHandler()

    def test_call(self):
        self.handlerCalled = self.globalCalled = False
        connection = object()
        dummyevent = DummyEvent()
        def handler(conn, event):
            self.assertIdentical(conn, connection)
            self.assertIdentical(event, dummyevent)
            self.handlerCalled = True
        def globalHandler(conn, event):
            self.assertIdentical(conn, connection)
            self.assertIdentical(event, dummyevent)
            self.globalCalled = True

        self.handler.addEventHandler(handler, DummyEvent)
        self.handler.addEventHandler(globalHandler)
        connection = object()
        self.handler(connection, dummyevent)
        self.failUnless(self.handlerCalled)
        self.failUnless(self.globalCalled)

    def test_callUnhandled(self):
        def handler(conn, event):
            self.fail()
        self.handler.addEventHandler(handler, DummyEvent)
        self.handler(None, AnotherDummyEvent())

    def test_clear(self):
        def handler(conn, event):
            self.fail()
        self.handler.addEventHandler(handler, DummyEvent)
        self.handler.addEventHandler(handler)
        self.handler.clear()
        self.handler(None, DummyEvent())

    def test_removeHandler(self):
        def handler(conn, event):
            self.fail()
        self.handler.addEventHandler(handler, DummyEvent)
        self.handler.addEventHandler(handler)
        self.handler.removeEventHandler(handler, DummyEvent)
        self.handler.removeEventHandler(handler)
        self.handler.clear()
        self.handler(None, DummyEvent())

    def test_exception(self):
        self.handlerCalled = False

        def failingHandler(conn, event):
            raise Exception()
        def handler(conn, event):
            self.handlerCalled = True

        self.handler.addEventHandler(failingHandler)
        self.handler.addEventHandler(handler)

        self.handler(None, DummyEvent())
        self.failUnless(self.handlerCalled)
        self.flushLoggedErrors()


class SynchronousHandlerTest(SimpleHandlerTest):

    def setUp(self):
        self.handler = event.SynchronousHandler()

    def test_locking(self):
        self.firstCalled = self.secondCalled = self.firstCalledAgain = False
        d1 = defer.Deferred()
        def handler(conn, event):
            if not self.firstCalled:
                self.firstCalled = True
                return d1
            else:
                self.firstCalledAgain = True
                return None

        d2 = defer.Deferred()
        def handler2(conn, event):
            if not self.secondCalled:
                self.secondCalled = True
                return d2
            else:
                return None
        self.handler.addEventHandler(handler2, DummyEvent)
        self.handler.addEventHandler(handler)
        self.handler(None, DummyEvent())
        self.handler(None, DummyEvent())
        self.failUnless(self.firstCalled)
        self.failIf(self.firstCalledAgain)
        self.failIf(self.secondCalled)
        d1.callback(None)
        self.failUnless(self.secondCalled)
        self.failIf(self.firstCalledAgain)
        d2.callback(None)
        self.failUnless(self.firstCalledAgain)
